#!/usr/bin/env python3

import asyncio
import json
import sys
import os
import stat
import ssl
import subprocess
import time
import logging
import argparse
import websockets
import pathlib
import tempfile


class MPV:
    reader: asyncio.StreamReader = None
    writer: asyncio.StreamWriter = None

    PAUSE_COMMAND = '{"command": ["set_property", "pause", true]}\n'
    PLAY_COMMAND = '{"command": ["set_property", "pause", false]}\n'
    SET_PLAYBACK_TIME = '{"command": ["set_property", "playback-time", %f]}\n'
    GET_PLAYBACK_TIME = '{"command": ["get_property", "playback-time"]}\n'

    is_playing = False

    @staticmethod
    def is_running():
        return (
            subprocess.run(["pgrep", "mpv"], stdout=subprocess.DEVNULL).returncode == 0
        )

    def start(self, video_file, socket_path):
        process = subprocess.Popen(
            ["mpv", "--pause", f"--input-ipc-server={socket_path}", video_file],
            preexec_fn=os.setpgrp,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        logging.info("Started mpv")
        time.sleep(1)

        if process.poll():
            logging.error(f"mpv terminated with returncode: {process.returncode}")

    async def play(self):
        await self._send_msg(self.PLAY_COMMAND)

    async def pause(self):
        await self._send_msg(self.PAUSE_COMMAND)

    async def get_playback_time(self):
        answer = await self._send_msg(self.GET_PLAYBACK_TIME)
        return answer["data"]

    async def set_playback_time(self, time):
        await self._send_msg(self.SET_PLAYBACK_TIME % time)

    async def _send_msg(self, msg: str):
        logging.info(f"Sending {msg} to mpv")
        self.writer.write(msg.encode())
        await self.writer.drain()
        logging.info(f"Finished sending {msg} to mpv")
        while True:
            answer = await self.reader.readuntil()
            logging.info(f"Got answer from mpv: {answer}")
            answer = json.loads(answer.decode())

            if "event" in answer:
                if answer["event"] == "pause":
                    self.is_playing = False
                elif answer["event"] == "unpause":
                    self.is_playing = True
                else:
                    logging.warning(f"Ignoring unsupported mpv event: {answer}.")
            elif "error" in answer:
                if answer["error"] != "success":
                    logging.error(f"mpv didn't like the command: {answer}.")
                return answer

    async def connect(self, socket_path):
        if not is_socket(socket_path):
            logging.error(
                f"<{socket_path}> is not a socket. It seems like mpv did not start \
    correctly or terminated prematurely"
            )
        self.reader, self.writer = await asyncio.open_unix_connection(socket_path)
        logging.info(f"Connected to mpv socket at <{socket_path}>")


player = MPV()


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="""\
Connect to the server and start a mpv instance if one is not already running. \
Relays command from the server to the mpv instance, but sanitizes the commands.""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "video_file",
        help="Path to the video file that you want to watch with your friends.",
    )

    parser.add_argument(
        "--hostname", help="The hostname of the server.", default="www.mpvsync.de"
    )

    parser.add_argument(
        "--port", help="The port of the server.", default=443, type=int
    )

    parser.add_argument("--ca-file", help="Path to the certicate file.", required=False)

    parser.add_argument(
        "--mpv-socket",
        help="Don't start mpv, but connect directly to the given mpv socket",
    )

    parser.add_argument(
        "--start-mpv",
        help="""Always start a new mpv instance. Otherwise checks if mpv is already \
running""",
        action="store_true",
    )

    parser.add_argument(
        "--ignore-cert",
        help="""Disable SSL hostname verification and ignore certificates.""",
        action="store_true",
    )

    parser.add_argument("--group", help="The group name.", default="")
    parser.add_argument("-v", "--verbose", action="store_true")

    args = parser.parse_args()

    logging_format = "%(asctime)s %(levelname)s %(message)s"
    if args.verbose:
        logging.basicConfig(level=logging.INFO, format=logging_format)
    else:
        logging.basicConfig(level=logging.WARNING, format=logging_format)

    video_file = args.video_file
    if not (os.path.isfile(video_file) and os.access(video_file, os.R_OK)):
        parser.error(
            f"Unable to access {video_file}. Make sure it exists and is accessible."
        )

    if args.start_mpv:
        args.do_start_mpv = True
    elif args.mpv_socket is not None:
        args.do_start_mpv = False
    else:
        args.do_start_mpv = not MPV.is_running()
    del args.start_mpv

    if args.mpv_socket is None:
        if "XDG_RUNTIME_DIR" in os.environ:
            args.tmp_dir = None
            xdg_runtime_dir = os.environ["XDG_RUNTIME_DIR"]
        else:
            args.tmp_dir = tempfile.TemporaryDirectory(dir="/tmp", prefix="mpvsync-")
            xdg_runtime_dir = args.tmp_dir.name

        args.mpv_socket = pathlib.Path(xdg_runtime_dir) / "mpvsocket"
    elif not is_socket(args.mpv_socket):
        parser.error(
            f"""The given socket path {args.mpv_socket} is not a UNIX socket"""
        )

    return args


def main():
    try:
        global player
        args = parse_arguments()

        if args.do_start_mpv:
            player.start(args.video_file, args.mpv_socket)

        async def async_main(args):
            await player.connect(args.mpv_socket)
            await handle_server(args.hostname, args.port, args.group, args.ignore_cert)

        asyncio.run(async_main(args))
    except KeyboardInterrupt:
        logging.info("Got KeyboardInterrupt. Terminating gracefully.")
        if args.tmp_dir is not None:
            args.tmp_dir.cleanup()
        sys.exit(0)


async def handle_server(hostname, port, group, ignore_cert):
    ssl_context = ssl.create_default_context()
    if ignore_cert:
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

    uri = f"wss://{hostname}/api/ws/{group}"

    while True:
        try:
            async with websockets.connect(uri, ssl=ssl_context) as ws:
                logging.info("Successfully connected to server")
                while True:
                    command_message = await ws.recv()
                    logging.info(f"Received command: {command_message}")
                    answer = await handle_command(json.loads(command_message))
                    if answer:
                        await ws.send(answer)

        except ConnectionRefusedError as e:
            logging.warning(
                f"Connection refused by server. {e}. Is the server running?"
            )
        except websockets.exceptions.ConnectionClosedError as e:
            logging.warning(f"Connection closed by server. {e}. Is the server running?")

        await asyncio.sleep(2)


async def handle_command(cmd):
    if cmd["command"] == "play":
        await player.play()
    elif cmd["command"] == "pause":
        await player.pause()
    elif cmd["command"] == "get-playback-time":
        playback_time = await player.get_playback_time()
        return '{"playback_time": %s}' % playback_time
    elif cmd["command"] == "set-playback-time":
        time_stamp = cmd["data"]
        if type(time_stamp) is int:
            time_stamp = float(time_stamp)
        if type(time_stamp) is not float:
            logging.warning(
                f"Received set-playback-time with a non-number: {time_stamp}. Ignoring..."
            )
        else:
            await player.set_playback_time(cmd["data"])
    else:
        logging.warning(f"Ignoring unknown command: {cmd}")


def is_socket(path):
    mode = os.stat(path).st_mode
    file_type = stat.S_IFMT(mode)
    return (file_type & stat.S_IFSOCK) == stat.S_IFSOCK


if __name__ == "__main__":
    main()
