import React, { RefObject } from "react";
import ReactPlayer from "react-player";
import { RootState } from "../store/store";
import { connect } from "react-redux";

import {
  selectPlaying,
  selectSeconds,
  setPlaySeconds,
  setPlayState,
  PlayerPlayingAction,
  PlayerSecondsAction,
} from "./PlayerSlice";
import { selectSubtibleFile, selectUrl } from "../url/UrlSlice";

interface StateProps {
  playing: boolean;
  seconds: number;
  url: string;
  subtitleFile: string;
}

interface DispatchProps {
  play: () => void;
  pause: () => void;
  setSeconds: (value: number) => void;
}

export type Props = StateProps & DispatchProps;

class Player extends React.Component<Props> {
  playerref: RefObject<ReactPlayer>;

  constructor(props: Props) {
    super(props);
    this.playerref = React.createRef<ReactPlayer>();
    this.handleProgress = this.handleProgress.bind(this);
    this.handleSeek = this.handleSeek.bind(this);
  }

  handleProgress(state: {
    played: number;
    playedSeconds: number;
    loaded: number;
    loadedSeconds: number;
  }): void {
    //if(state.playedSeconds <= 0.01) return;
    if (!ReactPlayer.canPlay(this.props.url)) return;
    this.props.setSeconds(state.playedSeconds);
  }

  handleSeek(seconds: number): void {
    if (!ReactPlayer.canPlay(this.props.url)) return;
    this.props.setSeconds(seconds);
  }

  async componentDidUpdate() {
    if (this.playerref.current === null) return;

    if (!ReactPlayer.canPlay(this.props.url)) return;

    // 50ms difference is fine, dont resync
    if (
      Math.abs(this.playerref.current.getCurrentTime() - this.props.seconds) <=
      0.05
    ) {
      return;
    }

    // Enough difference now seek to it!
    this.playerref.current.seekTo(this.props.seconds);
  }

  render() {
    let url = this.props.url || "?";
    let subs = this.props.subtitleFile || "?";

    return (
      <div style={{ position: "relative", paddingTop: "56.25%" }}>
        <ReactPlayer
          key={url + subs}
          url={url}
          style={{
            position: "absolute",
            top: 0,
            left: 0,
          }}
          width="100%"
          height="100%"
          progressInterval={10}
          controls={true}
          playing={this.props.playing}
          onProgress={this.handleProgress}
          onPlay={this.props.play}
          onPause={this.props.pause}
          onSeek={this.handleSeek}
          ref={this.playerref}
          config={{
            file: {
              tracks: [
                {
                  kind: "subtitles",
                  src: subs,
                  srcLang: "en",
                  default: true,
                  label: "en",
                },
              ],
            },
          }}
        />
      </div>
    );
  }
}

const mapStateToProps = (state: RootState): StateProps => ({
  playing: selectPlaying(state),
  seconds: selectSeconds(state),
  url: selectUrl(state),
  subtitleFile: selectSubtibleFile(state),
});

type Dispatch = (action: PlayerPlayingAction | PlayerSecondsAction) => {};
const mapDispatchToProps = (dispatch: Dispatch): DispatchProps => ({
  play(): void {
    dispatch(setPlayState(true));
  },
  pause(): void {
    dispatch(setPlayState(false));
  },
  setSeconds(value: number): void {
    dispatch(setPlaySeconds(value));
  },
});

export default connect(mapStateToProps, mapDispatchToProps)(Player);
