import React from "react";
import { RootState } from "../store/store";
import { connect } from "react-redux";

import "./style.css";

import {
  UrlAction,
  selectUrl,
  setUrl,
  setSubtitleFile,
  selectSubtibleFile,
} from "./UrlSlice";
import { Button, Grid } from "@material-ui/core";
import { createFFmpeg, fetchFile } from "@ffmpeg/ffmpeg";
import UploadButton from "./UploadButton";
import UrlTextfield from "./UrlTextfield";

interface StateProps {
  url: string;
  subtitle: string;
}

interface DispatchProps {
  dispatchUrl: (url: string) => void;
  dispatchSubtitleFile: (file: string) => void;
}

export type Props = StateProps & DispatchProps;

class UrlBox extends React.Component<Props> {
  constructor(props: Props) {
    super(props);
    this.handleURLChange = this.handleURLChange.bind(this);
    this.handleSubtitleChange = this.handleSubtitleChange.bind(this);
    this.handleServerClick = this.handleServerClick.bind(this);
    this.handleInput = this.handleInput.bind(this);
    this.handleSubtitleClick = this.handleSubtitleClick.bind(this);
  }

  handleURLChange(
    event: React.ChangeEvent<HTMLInputElement>
  ): void | undefined {
    this.props.dispatchUrl(event.target.value);
  }

  handleSubtitleChange(
    event: React.ChangeEvent<HTMLInputElement>
  ): void | undefined {
    this.props.dispatchSubtitleFile(event.target.value);
  }

  handleServerClick(): void | undefined {
    this.props.dispatchUrl("/video/index.mp4");
    this.props.dispatchSubtitleFile("/video/index.vtt");
  }

  async handleSubtitleClick(): Promise<void | undefined> {
    try {
      //let buffer = new Uint8Array(readFileSync(this.props.url))
      const ffmpeg = createFFmpeg({
        log: true,
      });
      await ffmpeg.load();
      this.props.dispatchSubtitleFile("Extracting...");

      // load file in browser
      ffmpeg.FS("writeFile", "video.avi", await fetchFile(this.props.url));
      // convert to vtt file
      await ffmpeg.run("-i", "video.avi", "outfile.vtt");
      // free memory from video file
      ffmpeg.FS("unlink", "video.avi");

      // create blob-url to read from
      let data: Uint8Array = ffmpeg.FS("readFile", "outfile.vtt");
      let blob = new Blob([data]);
      let u = URL.createObjectURL(blob);

      this.props.dispatchSubtitleFile(u);
    } catch (error) {
      this.props.dispatchSubtitleFile("Error...");
      console.log(error);
    }
  }

  handleInput(
    event: React.ChangeEvent<HTMLInputElement>,
    dispatchFunction: (f: string) => void
  ): void | undefined {
    if (event.target.files === null || event.target.files.length < 1) {
      return;
    }

    const selectedFile = event.target.files[0];
    console.log(selectedFile);
    event.target.files = null; // needed for proper update again
    let url = URL.createObjectURL(selectedFile);
    dispatchFunction(url);
  }

  render() {
    return (
      <Grid container>
        <Grid item xs={8}>
          <Grid container spacing={1}>
            <Grid item xs={9}>
              <UrlTextfield
                label="Video Source"
                value={this.props.url}
                onChangeFunc={this.handleURLChange}
              />
            </Grid>
            <Grid item xs={3}>
              <UploadButton
                onChangeFunc={(event: React.ChangeEvent<HTMLInputElement>) =>
                  this.handleInput(event, this.props.dispatchUrl)
                }
              />
            </Grid>

            <Grid item xs={9}>
              <UrlTextfield
                label="Subtitle Source"
                value={this.props.subtitle}
                onChangeFunc={this.handleSubtitleChange}
              />
            </Grid>
            <Grid item xs={3}>
              <UploadButton
                onChangeFunc={(event: React.ChangeEvent<HTMLInputElement>) =>
                  this.handleInput(event, this.props.dispatchSubtitleFile)
                }
              />
            </Grid>
          </Grid>
        </Grid>

        <Grid item xs={4}>
          <Grid container justify="flex-end" spacing={1}>
            <Grid item xs={10}>
              <Button
                className="urlElemHeight"
                size="large"
                variant="outlined"
                fullWidth
                onClick={this.handleServerClick}
              >
                Stream from server
              </Button>
            </Grid>
            <Grid item xs={10}>
              <Button
                className="urlElemHeight"
                size="large"
                fullWidth
                variant="outlined"
                onClick={this.handleSubtitleClick}
              >
                Extract subtitles (beta)
              </Button>
            </Grid>
          </Grid>
        </Grid>
      </Grid>
    );
  }
}

//<Button variant="outlined" color="primary" onClick={this.handleLocalClick}>Stream from local Stream from server</Button>
const mapStateToProps = (state: RootState): StateProps => ({
  url: selectUrl(state),
  subtitle: selectSubtibleFile(state),
});

type Dispatch = (action: UrlAction) => {};
const mapDispatchToProps = (dispatch: Dispatch): DispatchProps => ({
  dispatchUrl(url: string): void {
    dispatch(setUrl(url));
  },
  dispatchSubtitleFile(file: string): void {
    dispatch(setSubtitleFile(file));
  },
});

export default connect(mapStateToProps, mapDispatchToProps)(UrlBox);
